"""
*******
Configs
*******

**Configs** is an INI configuration parsing package, written for humans.

Basic usage
===========
Load a config file::

    >>> import configs
    >>> c = configs.load('sample.conf')
    >>> c['general']
    {'foo': 'baz'}

Load a config file with a fallback config file (with default values)::

    >>> fc = configs.load('sample.conf', fallback_file='default.conf')
    >>> fc['general']['spam']
    eggs

See the full documentation at `configs.rtfd.org <http://configs.rtfd.org>`_.

The repo is at `bitbucket.org/moigagoo/configs <https://bitbucket.org/moigagoo/configs>`_.
"""

__title__ = 'configs'
__version__ = '1.8.1'
__author__ = 'Konstantin Molchanov'
__license__ = 'MIT'

from .section import Section
from .config import Config
from .api import load



"""
**************
configs.config
**************

This module contains the :class:`Config <Config>` class.
"""

import re

from .section import Section

class Config:
    """Parsed configuration.

    Config instance includes a list of :class:`Section <Section>` instances.
    """

    _comment = re.compile('^\s*;.*$')
    _header = re.compile('^\s*\[\s*(?P<section>\w+)\s*\]\s*$')
    _dict_item = re.compile('^\s*(?P<key>\w+)\s*\=\s*(?P<value>.+)\s*$')
    _list_item = re.compile('^\s*(?P<value>.+)\s*$')

    def __init__(self, config_file=None, fallback=None):
        if fallback:
            self.sections = fallback.sections
        else:
            self.sections = {}
            self._add_section('root')

        if config_file:
            self.load(config_file)

    def get(self):
        """Gets all section items."""

        return {section: self.sections[section].get() for section in self.sections}

    def load(self, config_file):
        """Parse an INI configuration file.
        
        :param config_file: configuration file to be loaded.
        
        :returns: :class:Config instance.
        """

        current_section = None

        with open(config_file, 'r') as f:
            for line in f.readlines():
                comment_match = re.match(self._comment, line)
                if comment_match:
                    continue

                header_match = re.match(self._header, line)
                if header_match:
                    current_section = header_match.group('section')
                    if not current_section in self.sections:
                        self._add_section(current_section)

                    continue

                dict_item_match = re.match(self._dict_item, line)
                if dict_item_match:
                    key, value = dict_item_match.group('key'), dict_item_match.group('value')

                    if current_section:
                        self._add_dict_prop_to_section(key, value, current_section)
                    else:
                        self._add_dict_prop_to_section(key, value)

                    continue

                list_item_match = re.match(self._list_item, line)
                if list_item_match:
                    value = list_item_match.group('value')
                    if current_section:
                        self._add_list_prop_to_section(value, current_section)
                    else:
                        self._add_list_prop_to_section(value)

                    continue

    def _add_section(self, name):
        """Adds an empty section with the given name.
        
        :param name: new section name.        
        """

        self.sections[name] = Section()

    def _add_dict_prop_to_section(self, key, value, section='root'):
        """Adds a key-value item to the given section.
        
        :param key: new item key.
        :param value: new item value.
        :param section: (optional) section name (``root`` by default).
        """

        if section in self.sections:
            self.sections[section]._add_dict_prop(key, value)
        else:
            raise KeyError

    def _add_list_prop_to_section(self, value, section='root'):
        """Adds a flag value to the given section.
        
        :param value: new item value.
        :param section: (optional) section name (``root`` by default).
        """

        if section in self.sections:
            self.sections[section]._add_list_prop(value)
        else:
            raise KeyError

    def __repr__(self):
        return str(self.get())

    def __str__(self):
        return str(self.get())

    def __getitem__(self, key):
        if key in self.sections:
            return self.sections[key]
        else:
            try:
                return self.sections['root'][key]
            except KeyError:
                pass

        return None

    def __eq__(self, other):
        return self.sections == other.sections


"""
***************
configs.section
***************

This module contains the :class:`Section <Section>` class.
"""

class Section:
    """INI configuration section.

    A Section instance stores both key-value and flag items, in ``dict_props`` and ``list_props`` attributes respectively.

    It is possible to iterate over a section; flag values are listed first, then key-value items.
    """
    
    def __init__(self):
        self.dict_props = {}
        self.list_props = []

    def get(self):
        """Gets section values.

        If section contains only flag values, a list is returned.

        If section contains only key-value items, a dictionary is returned.

        If section contains both flag and key-value items, a tuple of both is returned.
        """
        
        if self.list_props and self.dict_props:
            return self.list_props, self.dict_props

        return self.list_props or self.dict_props or None

    def _add_dict_prop(self, key, value):
        """Adds a key-value item to section."""

        try:
            value = float(value)
            value = int(value)
        except ValueError:
            pass

        self.dict_props[key] = value

    def _add_list_prop(self, value):
        """Adds a flag value to  section."""

        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                pass

        self.list_props.append(value)

    def __repr__(self):
        return str(self.get())

    def __str__(self):
        return str(self.get())

    def __iter__(self):
        for list_prop in self.list_props:
            yield list_prop

        for dict_prop in self.dict_props:
            yield dict_prop

    def __getitem__(self, key):
        try:
            return self.dict_props[key]
        except KeyError:
            pass

        try:
            return self.list_props[key]
        except (KeyError, TypeError):
            pass

        return None

    def __eq__(self, other):
        return self.dict_props == other.dict_props and self.list_props == other.list_props


"""
***********
configs.api
***********

This module implements the configs API.
"""

from .config import Config

def load(config_file, fallback_file=None):
    """Constructs are returns a :class:`Config <Config>` instance.

    :param config_file: configuration file to be parsed
    :param fallback_file: (optional) fallback configuration file with default values to be used if missing in the ``config_file``

    Usage::

        >>> import configs

        >>> fc = configs.load('sample.conf', fallback_file='default.conf')

        >>> fc['general']['spam']
        eggs
    """

    if fallback_file:
        return Config(config_file, Config(fallback_file))
    else:
        return Config(config_file)




from distutils.core import setup

import configs

setup(
    name=configs.__title__,
    version=configs.__version__,
    author=configs.__author__,
    description='Configuration for humans',
    long_description=open('README.rst').read(),
    author_email='moigagoo@myopera.com',
    url='https://bitbucket.org/moigagoo/configs',
    packages=['configs'],
    package_dir={'configs': 'configs'},
    package_data={'configs': ['*.conf']},
    license=configs.__license__
)
